@@ -1,58 +1,55 @@
 /*
  * pg_reorg: lib/reorg.c
  *
- * Copyright (c) 2008-2009, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
- */
-
-/**
- * @brief Core Modules
+ * Copyright (c) 2008-2010, NIPPON TELEGRAPH AND TELEPHONE CORPORATION
  */
 
 #include "postgres.h"
+
+#include <unistd.h>
+
+#include "access/genam.h"
 #include "access/transam.h"
 #include "access/xact.h"
 #include "catalog/dependency.h"
 #include "catalog/indexing.h"
 #include "catalog/namespace.h"
 #include "catalog/pg_namespace.h"
+#include "catalog/pg_opclass.h"
 #include "catalog/pg_type.h"
 #include "commands/tablecmds.h"
 #include "commands/trigger.h"
-#include "executor/spi.h"
 #include "miscadmin.h"
 #include "utils/builtins.h"
 #include "utils/lsyscache.h"
 #include "utils/relcache.h"
 #include "utils/syscache.h"
 
+#include "pgut/pgut-spi.h"
+#include "pgut/pgut-be.h"
+
 PG_MODULE_MAGIC;
 
-#if PG_VERSION_NUM < 80300
-#define SET_VARSIZE(PTR, len)	(VARATT_SIZEP((PTR)) = (len))
-#define PGDLLIMPORT				DLLIMPORT
-typedef void *SPIPlanPtr;
-#endif
-
-Datum reorg_trigger(PG_FUNCTION_ARGS);
-Datum reorg_apply(PG_FUNCTION_ARGS);
-Datum reorg_get_index_keys(PG_FUNCTION_ARGS);
-Datum reorg_indexdef(PG_FUNCTION_ARGS);
-Datum reorg_swap(PG_FUNCTION_ARGS);
-Datum reorg_drop(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT reorg_version(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT reorg_trigger(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT reorg_apply(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT reorg_get_index_keys(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT reorg_indexdef(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT reorg_swap(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT reorg_drop(PG_FUNCTION_ARGS);
+extern Datum PGUT_EXPORT reorg_disable_autovacuum(PG_FUNCTION_ARGS);
 
+PG_FUNCTION_INFO_V1(reorg_version);
 PG_FUNCTION_INFO_V1(reorg_trigger);
 PG_FUNCTION_INFO_V1(reorg_apply);
 PG_FUNCTION_INFO_V1(reorg_get_index_keys);
 PG_FUNCTION_INFO_V1(reorg_indexdef);
 PG_FUNCTION_INFO_V1(reorg_swap);
 PG_FUNCTION_INFO_V1(reorg_drop);
+PG_FUNCTION_INFO_V1(reorg_disable_autovacuum);
 
 static void	reorg_init(void);
 static SPIPlanPtr reorg_prepare(const char *src, int nargs, Oid *argtypes);
-static void reorg_execp(SPIPlanPtr plan, Datum *values, const char *nulls, int expected);
-static void reorg_execf(int expexted, const char *format, ...)
-__attribute__((format(printf, 2, 3)));
-static void reorg_execd(const char *src, int nargs, Oid *argtypes, Datum *values, const char *nulls, int expected);
 static const char *get_quoted_relname(Oid oid);
 static const char *get_quoted_nspname(Oid oid);
 static void swap_heap_or_index_files(Oid r1, Oid r2);
@@ -60,6 +57,9 @@ static void swap_heap_or_index_files(Oid r1, Oid r2);
 #define copy_tuple(tuple, desc) \
 	PointerGetDatum(SPI_returntuple((tuple), (desc)))
 
+#define IsToken(c) \
+	(IS_HIGHBIT_SET((c)) || isalnum((unsigned char) (c)) || (c) == '_')
+
 /* check access authority */
 static void
 must_be_superuser(const char *func)
@@ -69,13 +69,15 @@ must_be_superuser(const char *func)
 }
 
 #if PG_VERSION_NUM < 80400
-static int SPI_execute_with_args(const char *src, int nargs, Oid *argtypes,
-					  Datum *values, const char *nulls,
-					  bool read_only, long tcount);
-static text *cstring_to_text(const char * s);
 static void RenameRelationInternal(Oid myrelid, const char *newrelname, Oid namespaceId);
 #endif
 
+Datum
+reorg_version(PG_FUNCTION_ARGS)
+{
+	return CStringGetTextDatum("pg_reorg 1.1.5");
+}
+
 /**
  * @fn      Datum reorg_trigger(PG_FUNCTION_ARGS)
  * @brief   Insert a operation log into log-table.
@@ -91,7 +93,7 @@ reorg_trigger(PG_FUNCTION_ARGS)
 	TupleDesc		desc;
 	HeapTuple		tuple;
 	Datum			values[2];
-	char			nulls[2] = { ' ', ' ' };
+	bool			nulls[2] = { 0, 0 };
 	Oid				argtypes[2];
 	const char	   *sql;
 
@@ -117,7 +119,7 @@ reorg_trigger(PG_FUNCTION_ARGS)
 	{
 		/* INSERT: (NULL, newtup) */
 		tuple = trigdata->tg_trigtuple;
-		nulls[0] = 'n';
+		nulls[0] = true;
 		values[1] = copy_tuple(tuple, desc);
 	}
 	else if (TRIGGER_FIRED_BY_DELETE(trigdata->tg_event))
@@ -125,7 +127,7 @@ reorg_trigger(PG_FUNCTION_ARGS)
 		/* DELETE: (oldtup, NULL) */
 		tuple = trigdata->tg_trigtuple;
 		values[0] = copy_tuple(tuple, desc);
-		nulls[1] = 'n';
+		nulls[1] = true;
 	}
 	else
 	{
@@ -136,7 +138,7 @@ reorg_trigger(PG_FUNCTION_ARGS)
 	}
 
 	/* INSERT INTO reorg.log VALUES ($1, $2) */
-	reorg_execd(sql, 2, argtypes, values, nulls, SPI_OK_INSERT);
+	execute_with_args(SPI_OK_INSERT, sql, 2, argtypes, values, nulls);
 
 	SPI_finish();
 
@@ -160,7 +162,7 @@ reorg_trigger(PG_FUNCTION_ARGS)
 Datum
 reorg_apply(PG_FUNCTION_ARGS)
 {
-#define NUMBER_OF_PROCESSING	1000
+#define DEFAULT_PEEK_COUNT	1000
 
 	const char *sql_peek = PG_GETARG_CSTRING(0);
 	const char *sql_insert = PG_GETARG_CSTRING(1);
@@ -175,13 +177,9 @@ reorg_apply(PG_FUNCTION_ARGS)
 	SPIPlanPtr		plan_update = NULL;
 	SPIPlanPtr		plan_pop = NULL;
 	uint32			n, i;
-	TupleDesc		desc;
-	Oid				argtypes[3];	/* id, pk, row */
-	Datum			values[3];		/* id, pk, row */
-	char			nulls[3];		/* id, pk, row */
 	Oid				argtypes_peek[1] = { INT4OID };
 	Datum			values_peek[1];
-	char			nulls_peek[1] = { ' ' };
+	bool			nulls_peek[1] = { 0 };
 
 	/* authority check */
 	must_be_superuser("reorg_apply");
@@ -196,14 +194,18 @@ reorg_apply(PG_FUNCTION_ARGS)
 	{
 		int				ntuples;
 		SPITupleTable  *tuptable;
+		TupleDesc		desc;
+		Oid				argtypes[3];	/* id, pk, row */
+		Datum			values[3];		/* id, pk, row */
+		bool			nulls[3];		/* id, pk, row */
 
 		/* peek tuple in log */
 		if (count == 0)
-			values_peek[0] = Int32GetDatum(NUMBER_OF_PROCESSING);
+			values_peek[0] = Int32GetDatum(DEFAULT_PEEK_COUNT);
 		else
-			values_peek[0] = Int32GetDatum(Min(count - n, NUMBER_OF_PROCESSING));
-		
-		reorg_execp(plan_peek, values_peek, nulls_peek, SPI_OK_SELECT);
+			values_peek[0] = Int32GetDatum(Min(count - n, DEFAULT_PEEK_COUNT));
+
+		execute_plan(SPI_OK_SELECT, plan_peek, values_peek, nulls_peek);
 		if (SPI_processed <= 0)
 			break;
 
@@ -218,43 +220,39 @@ reorg_apply(PG_FUNCTION_ARGS)
 		for (i = 0; i < ntuples; i++, n++)
 		{
 			HeapTuple	tuple;
-			bool		isnull;
-		
+
 			tuple = tuptable->vals[i];
-			values[0] = SPI_getbinval(tuple, desc, 1, &isnull);
-			nulls[0] = ' ';
-			values[1] = SPI_getbinval(tuple, desc, 2, &isnull);
-			nulls[1] = (isnull ? 'n' : ' ');
-			values[2] = SPI_getbinval(tuple, desc, 3, &isnull);
-			nulls[2] = (isnull ? 'n' : ' ');
-		
-			if (nulls[1] == 'n')
+			values[0] = SPI_getbinval(tuple, desc, 1, &nulls[0]);
+			values[1] = SPI_getbinval(tuple, desc, 2, &nulls[1]);
+			values[2] = SPI_getbinval(tuple, desc, 3, &nulls[2]);
+
+			if (nulls[1])
 			{
 				/* INSERT */
 				if (plan_insert == NULL)
 					plan_insert = reorg_prepare(sql_insert, 1, &argtypes[2]);
-				reorg_execp(plan_insert, &values[2], &nulls[2], SPI_OK_INSERT);
+				execute_plan(SPI_OK_INSERT, plan_insert, &values[2], &nulls[2]);
 			}
-			else if (nulls[2] == 'n')
+			else if (nulls[2])
 			{
 				/* DELETE */
 				if (plan_delete == NULL)
 					plan_delete = reorg_prepare(sql_delete, 1, &argtypes[1]);
-				reorg_execp(plan_delete, &values[1], &nulls[1], SPI_OK_DELETE);
+				execute_plan(SPI_OK_DELETE, plan_delete, &values[1], &nulls[1]);
 			}
 			else
 			{
 				/* UPDATE */
 				if (plan_update == NULL)
 					plan_update = reorg_prepare(sql_update, 2, &argtypes[1]);
-				reorg_execp(plan_update, &values[1], &nulls[1], SPI_OK_UPDATE);
+				execute_plan(SPI_OK_UPDATE, plan_update, &values[1], &nulls[1]);
 			}
 		}
 
 		/* delete tuple in log */
 		if (plan_pop == NULL)
 			plan_pop = reorg_prepare(sql_pop, 1, argtypes);
-		reorg_execp(plan_pop, values, nulls, SPI_OK_DELETE);
+		execute_plan(SPI_OK_DELETE, plan_pop, values, nulls);
 
 		SPI_freetuptable(tuptable);
 	}
@@ -265,7 +263,7 @@ reorg_apply(PG_FUNCTION_ARGS)
 }
 
 /*
- * Deparsed create index sql. You can rebuild sql using 
+ * Parsed CREATE INDEX statement. You can rebuild sql using 
  * sprintf(buf, "%s %s ON %s USING %s (%s)%s",
  *		create, index, table type, columns, options)
  */
@@ -280,7 +278,7 @@ typedef struct IndexDef
 } IndexDef;
 
 static char *
-generate_relation_name(Oid relid)
+get_relation_name(Oid relid)
 {
 	Oid		nsp = get_rel_namespace(relid);
 	char   *nspname;
@@ -297,7 +295,7 @@ generate_relation_name(Oid relid)
 static char *
 parse_error(Oid index)
 {
-	elog(ERROR, "Unexpected indexdef: %s", pg_get_indexdef_string(index));
+	elog(ERROR, "unexpected index definition: %s", pg_get_indexdef_string(index));
 	return NULL;
 }
 
@@ -320,24 +318,49 @@ skip_const(Oid index, char *sql, const char *arg1, const char *arg2)
 static char *
 skip_ident(Oid index, char *sql)
 {
-	sql = strchr(sql, ' ');
-	if (sql)
+	while (*sql && isspace((unsigned char) *sql))
+		sql++;
+
+	if (*sql == '"')
 	{
+		sql++;
+		for (;;)
+		{
+			char *end = strchr(sql, '"');
+			if (end == NULL)
+				return parse_error(index);
+			else if (end[1] != '"')
+			{
+				end[1] = '\0';
+				return end + 2;
+			}
+			else	/* escaped quote ("") */
+				sql = end + 2;
+		}
+	}
+	else
+	{
+		while (*sql && IsToken(*sql))
+			sql++;
 		*sql = '\0';
-		return sql + 2;
+		return sql + 1;
 	}
 
 	/* error */
 	return parse_error(index);
 }
 
+/*
+ * Skip until 'end' character found. The 'end' character is replaced with \0.
+ * Returns the next character of the 'end', or NULL if 'end' is not found.
+ */
 static char *
-skip_columns(Oid index, char *sql)
+skip_until(Oid index, char *sql, char end)
 {
 	char	instr = 0;
-	int		nopen = 1;
+	int		nopen = 0;
 
-	for (; *sql && nopen > 0; sql++)
+	for (; *sql && (nopen > 0 || instr != 0 || *sql != end); sql++)
 	{
 		if (instr)
 		{
@@ -349,9 +372,7 @@ skip_columns(Oid index, char *sql)
 					instr = 0;
 			}
 			else if (sql[0] == '\\')
-			{
-				sql++;	// next char is always string
-			}
+				sql++;	/* next char is always string */
 		}
 		else
 		{
@@ -371,10 +392,15 @@ skip_columns(Oid index, char *sql)
 		}
 	}
 
-	if (nopen == 0)
+	if (nopen == 0 && instr == 0)
 	{
-		sql[-1] = '\0';
-		return sql;
+		if (*sql)
+		{
+			*sql = '\0';
+			return sql + 1;
+		}
+		else
+			return NULL;
 	}
 
 	/* error */
@@ -386,7 +412,7 @@ parse_indexdef(IndexDef *stmt, Oid index, Oid table)
 {
 	char *sql = pg_get_indexdef_string(index);
 	const char *idxname = get_quoted_relname(index);
-	const char *tblname = generate_relation_name(table);
+	const char *tblname = get_relation_name(table);
 
 	/* CREATE [UNIQUE] INDEX */
 	stmt->create = sql;
@@ -402,11 +428,15 @@ parse_indexdef(IndexDef *stmt, Oid index, Oid table)
 	/* USING */
 	sql = skip_const(index, sql, "USING", NULL);
 	/* type */
-	stmt->type= sql;
+	stmt->type = sql;
 	sql = skip_ident(index, sql);
 	/* (columns) */
+	if ((sql = strchr(sql, '(')) == NULL)
+		parse_error(index);
+	sql++;
 	stmt->columns = sql;
-	sql = skip_columns(index, sql);
+	if ((sql = skip_until(index, sql, ')')) == NULL)
+		parse_error(index);
 	/* options */
 	stmt->options = sql;
 }
@@ -420,6 +450,9 @@ parse_indexdef(IndexDef *stmt, Oid index, Oid table)
  * @param	index	Oid of target index.
  * @param	table	Oid of table of the index.
  * @retval			Create index DDL for temp table.
+ *
+ * FIXME: this function is named get_index_keys, but actually returns
+ * an expression for ORDER BY clause. get_order_by() might be a better name.
  */
 Datum
 reorg_get_index_keys(PG_FUNCTION_ARGS)
@@ -427,10 +460,94 @@ reorg_get_index_keys(PG_FUNCTION_ARGS)
 	Oid				index = PG_GETARG_OID(0);
 	Oid				table = PG_GETARG_OID(1);
 	IndexDef		stmt;
+	char		   *token;
+	char		   *next;
+	StringInfoData	str;
+	Relation		indexRel = NULL;
+	int				nattr;
 
 	parse_indexdef(&stmt, index, table);
+	elog(DEBUG2, "indexdef.create  = %s", stmt.create);
+	elog(DEBUG2, "indexdef.index   = %s", stmt.index);
+	elog(DEBUG2, "indexdef.table   = %s", stmt.table);
+	elog(DEBUG2, "indexdef.type    = %s", stmt.type);
+	elog(DEBUG2, "indexdef.columns = %s", stmt.columns);
+	elog(DEBUG2, "indexdef.options = %s", stmt.options);
 
-	PG_RETURN_TEXT_P(cstring_to_text(stmt.columns));
+	/*
+	 * FIXME: this is very unreliable implementation but I don't want to
+	 * re-implement customized versions of pg_get_indexdef_string...
+	 *
+	 * TODO: Support ASC/DESC and NULL FIRST/LAST.
+	 */
+
+	initStringInfo(&str);
+	for (nattr = 0, next = stmt.columns; next; nattr++)
+	{
+		char *opcname;
+
+		token = next;
+		while (isspace((unsigned char) *token))
+			token++;
+		next = skip_until(index, next, ',');
+		opcname = skip_until(index, token, ' ');
+		if (opcname)
+		{
+			/* lookup default operator name from operator class */
+
+			Oid				opclass;
+			Oid				oprid;
+			int16			strategy = BTLessStrategyNumber;
+#if PG_VERSION_NUM >= 80300
+			Oid				opcintype;
+			Oid				opfamily;
+			HeapTuple		tp;
+			Form_pg_opclass	opclassTup;
+#endif
+			
+			opclass = OpclassnameGetOpcid(BTREE_AM_OID, opcname);
+
+#if PG_VERSION_NUM >= 80300
+			/* Retrieve operator information. */
+			tp = SearchSysCache(CLAOID, ObjectIdGetDatum(opclass), 0, 0, 0);
+			if (!HeapTupleIsValid(tp))
+				elog(ERROR, "cache lookup failed for opclass %u", opclass);
+			opclassTup = (Form_pg_opclass) GETSTRUCT(tp);
+			opfamily = opclassTup->opcfamily;
+			opcintype = opclassTup->opcintype;
+			ReleaseSysCache(tp);
+
+			if (!OidIsValid(opcintype))
+			{
+				if (indexRel == NULL)
+					indexRel = index_open(index, NoLock);
+
+				opcintype = RelationGetDescr(indexRel)->attrs[nattr]->atttypid;
+			}
+
+			oprid = get_opfamily_member(opfamily, opcintype, opcintype, strategy);
+			if (!OidIsValid(oprid))
+				elog(ERROR, "missing operator %d(%u,%u) in opfamily %u",
+					 strategy, opcintype, opcintype, opfamily);
+#else
+			oprid = get_opclass_member(opclass, 0, strategy);
+			if (!OidIsValid(oprid))
+				elog(ERROR, "missing operator %d for %s", strategy, opcname);
+#endif
+
+				opcname[-1] = '\0';
+			appendStringInfo(&str, "%s USING %s", token, get_opname(oprid));
+		}
+		else
+			appendStringInfoString(&str, token);
+		if (next)
+			appendStringInfoString(&str, ", ");
+	}
+
+	if (indexRel != NULL)
+		index_close(indexRel, NoLock);
+
+	PG_RETURN_TEXT_P(cstring_to_text(str.data));
 }
 
 /**
@@ -467,6 +584,109 @@ getoid(HeapTuple tuple, TupleDesc desc, int column)
 	return isnull ? InvalidOid : DatumGetObjectId(datum);
 }
 
+static int16
+getint16(HeapTuple tuple, TupleDesc desc, int column)
+{
+	bool	isnull;
+	Datum	datum = SPI_getbinval(tuple, desc, column, &isnull);
+	return isnull ? 0 : DatumGetInt16(datum);
+}
+
+static void
+remove_dropped_columns_and_adjust_attnum(Oid oid, int16 natts1, int16 natts2)
+{
+	/* delete dropped columns */
+	execute_with_format(SPI_OK_DELETE,
+		"DELETE FROM pg_catalog.pg_attribute"
+		" WHERE attrelid = %u AND attisdropped",
+		oid);
+	if (SPI_processed != natts1 - natts2)
+		elog(ERROR, "cannot remove %d dropped columns (%u columns removed)",
+			natts2 - natts1, SPI_processed);
+
+	/* renumber attnum */
+#if PG_VERSION_NUM >= 90000
+	execute_with_format(SPI_OK_UPDATE,
+		"UPDATE pg_catalog.pg_attribute m"
+		"   SET attnum = (SELECT count(*) FROM pg_attribute a"
+		" WHERE m.attrelid = a.attrelid"
+		"   AND m.attnum >= a.attnum"
+		"   AND a.attnum > 0 AND NOT a.attisdropped)"
+		" WHERE attrelid = %u AND attnum > 0 AND NOT attisdropped",
+		oid);
+	if (SPI_processed != natts2)
+		elog(ERROR, "cannot update %d columns (%u columns updated)",
+			natts2, SPI_processed);
+#elif PG_VERSION_NUM >= 80300
+	execute_with_format(SPI_OK_UPDATE,
+		"UPDATE pg_catalog.pg_attribute"
+		"   SET attnum = (SELECT count(*) FROM pg_attribute a"
+		" WHERE pg_catalog.pg_attribute.attrelid = a.attrelid"
+		"   AND pg_catalog.pg_attribute.attnum >= a.attnum"
+		"   AND a.attnum > 0 AND NOT a.attisdropped)"
+		" WHERE attrelid = %u AND attnum > 0 AND NOT attisdropped",
+		oid);
+	if (SPI_processed != natts2)
+		elog(ERROR, "cannot update %d columns (%u columns updated)",
+			natts2, SPI_processed);
+#else
+	/*
+	 * Use count(*) in subquery because 8.2 doesn't support aggregates
+	 * in UPDATE SET.
+	 */
+	do
+	{
+		uint32			i;
+		uint32			ntuples;
+		SPITupleTable  *tuptable;
+		TupleDesc		desc;
+
+		execute_with_format(SPI_OK_SELECT,
+			"SELECT attnum FROM pg_catalog.pg_attribute"
+			" WHERE attrelid = %u AND attnum > 0 AND NOT attisdropped"
+			" ORDER BY attnum",
+			oid);
+		if (SPI_processed != natts2)
+			elog(ERROR, "number of columns should be %d (%d returned)",
+				natts2, SPI_processed);
+
+		ntuples = SPI_processed;
+		tuptable = SPI_tuptable;
+		desc = tuptable->tupdesc;
+
+		for (i = 0; i < ntuples; i++)
+		{
+			int		attnum;
+			int		count;
+
+			attnum = getint16(tuptable->vals[i], desc, 1);
+
+			execute_with_format(SPI_OK_SELECT,
+				"SELECT count(*)::smallint FROM pg_catalog.pg_attribute"
+				" WHERE attrelid = %u AND attnum > 0 AND attnum <= %d",
+				oid, attnum);
+			if (SPI_processed != 1)
+				elog(ERROR, "cannot adjust column %d", attnum);
+
+			count = getint16(SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1);
+
+			execute_with_format(SPI_OK_UPDATE,
+				"UPDATE pg_catalog.pg_attribute"
+				"   SET attnum = %d"
+				" WHERE attrelid = %u AND attnum = %d",
+				count, oid, attnum);
+			if (SPI_processed != 1)
+				elog(ERROR, "cannot update column %d", attnum);
+		}
+	} while(0);
+#endif
+
+	/* adjust attribute number of the table */
+	execute_with_format(SPI_OK_UPDATE,
+		"UPDATE pg_catalog.pg_class SET relnatts = %d WHERE oid = %u",
+		natts2, oid);
+}
+
 /**
  * @fn      Datum reorg_swap(PG_FUNCTION_ARGS)
  * @brief   Swapping relfilenode of tables and relation ids of toast tables
@@ -474,6 +694,8 @@ getoid(HeapTuple tuple, TupleDesc desc, int column)
  *
  * reorg_swap(oid, relname)
  *
+ * TODO: remove useless CommandCounterIncrement().
+ *
  * @param	oid		Oid of table of target.
  * @retval			None.
  */
@@ -484,7 +706,7 @@ reorg_swap(PG_FUNCTION_ARGS)
 	const char	   *relname = get_quoted_relname(oid);
 	const char	   *nspname = get_quoted_nspname(oid);
 	Oid 			argtypes[1] = { OIDOID };
-	char	 		nulls[1] = { ' ' };
+	bool	 		nulls[1] = { 0 };
 	Datum	 		values[1];
 	SPITupleTable  *tuptable;
 	TupleDesc		desc;
@@ -497,6 +719,10 @@ reorg_swap(PG_FUNCTION_ARGS)
 	Oid				oid2;
 	Oid				reltoastrelid2;
 	Oid				reltoastidxid2;
+	Oid				owner1;
+	Oid				owner2;
+	int16			natts1;
+	int16			natts2;
 
 	/* authority check */
 	must_be_superuser("reorg_swap");
@@ -506,14 +732,17 @@ reorg_swap(PG_FUNCTION_ARGS)
 
 	/* swap relfilenode and dependencies for tables. */
 	values[0] = ObjectIdGetDatum(oid);
-	reorg_execd(
-		"SELECT X.oid, X.reltoastrelid, TX.reltoastidxid,"
-		"       Y.oid, Y.reltoastrelid, TY.reltoastidxid"
-		"  FROM pg_class X LEFT JOIN pg_class TX ON X.reltoastrelid = TX.oid,"
-		"       pg_class Y LEFT JOIN pg_class TY ON Y.reltoastrelid = TY.oid"
+	execute_with_args(SPI_OK_SELECT,
+		"SELECT X.reltoastrelid, TX.reltoastidxid, X.relowner,"
+		"       Y.oid, Y.reltoastrelid, TY.reltoastidxid, Y.relowner,"
+		"       X.relnatts, Y.relnatts"
+		"  FROM pg_catalog.pg_class X LEFT JOIN pg_catalog.pg_class TX"
+		"         ON X.reltoastrelid = TX.oid,"
+		"       pg_catalog.pg_class Y LEFT JOIN pg_catalog.pg_class TY"
+		"         ON Y.reltoastrelid = TY.oid"
 		" WHERE X.oid = $1"
 		"   AND Y.oid = ('reorg.table_' || X.oid)::regclass",
-		1, argtypes, values, nulls, SPI_OK_SELECT);
+		1, argtypes, values, nulls);
 
 	tuptable = SPI_tuptable;
 	desc = tuptable->tupdesc;
@@ -524,34 +753,38 @@ reorg_swap(PG_FUNCTION_ARGS)
 
 	tuple = tuptable->vals[0];
 
-	reltoastrelid1 = getoid(tuple, desc, 2);
-	reltoastidxid1 = getoid(tuple, desc, 3);
+	reltoastrelid1 = getoid(tuple, desc, 1);
+	reltoastidxid1 = getoid(tuple, desc, 2);
+	owner1 = getoid(tuple, desc, 3);
 	oid2 = getoid(tuple, desc, 4);
 	reltoastrelid2 = getoid(tuple, desc, 5);
 	reltoastidxid2 = getoid(tuple, desc, 6);
+	owner2 = getoid(tuple, desc, 7);
+	natts1 = getint16(tuple, desc, 8);
+	natts2 = getint16(tuple, desc, 9);;
 
-	/* should be all-or-nothing */
-	if ((reltoastrelid1 == InvalidOid || reltoastidxid1 == InvalidOid ||
-		 reltoastrelid2 == InvalidOid || reltoastidxid2 == InvalidOid) &&
-		(reltoastrelid1 != InvalidOid || reltoastidxid1 != InvalidOid ||
-		 reltoastrelid2 != InvalidOid || reltoastidxid2 != InvalidOid))
+	/* change owner of new relation to original owner */
+	if (owner1 != owner2)
 	{
-		elog(ERROR, "reorg_swap : unexpected toast relations (T1=%u, I1=%u, T2=%u, I2=%u",
-			reltoastrelid1, reltoastidxid1, reltoastrelid2, reltoastidxid2);
+		ATExecChangeOwner(oid2, owner1, true);
+		CommandCounterIncrement();
 	}
 
+	/* swap tables. */
 	swap_heap_or_index_files(oid, oid2);
 	CommandCounterIncrement();
 
-	/* swap relfilenode and dependencies for indxes. */
+	/* swap indexes. */
 	values[0] = ObjectIdGetDatum(oid);
-	reorg_execd(
+	execute_with_args(SPI_OK_SELECT,
 		"SELECT X.oid, Y.oid"
-		"  FROM pg_index I, pg_class X, pg_class Y"
+		"  FROM pg_catalog.pg_index I,"
+		"       pg_catalog.pg_class X,"
+		"       pg_catalog.pg_class Y"
 		" WHERE I.indrelid = $1"
 		"   AND I.indexrelid = X.oid"
 		"   AND Y.oid = ('reorg.index_' || X.oid)::regclass",
-		1, argtypes, values, nulls, SPI_OK_SELECT);
+		1, argtypes, values, nulls);
 
 	tuptable = SPI_tuptable;
 	desc = tuptable->tupdesc;
@@ -559,15 +792,64 @@ reorg_swap(PG_FUNCTION_ARGS)
 
 	for (i = 0; i < records; i++)
 	{
+		Oid		idx1, idx2;
+
 		tuple = tuptable->vals[i];
-		swap_heap_or_index_files(
-			getoid(tuple, desc, 1),
-			getoid(tuple, desc, 2));
+		idx1 = getoid(tuple, desc, 1);
+		idx2 = getoid(tuple, desc, 2);
+		swap_heap_or_index_files(idx1, idx2);
+
+		/* adjust key attnum if the target table has dropped columns */
+		if (natts1 != natts2)
+		{
+#if PG_VERSION_NUM >= 90000
+			execute_with_format(SPI_OK_UPDATE,
+				"UPDATE pg_catalog.pg_index m SET indkey = n.indkey"
+				"  FROM pg_catalog.pg_index n"
+				" WHERE m.indexrelid = %u"
+				"   AND n.indexrelid = 'reorg.index_%u'::regclass",
+				idx1, idx1);
+#else
+			execute_with_format(SPI_OK_UPDATE,
+				"UPDATE pg_catalog.pg_index SET indkey = n.indkey"
+				"  FROM pg_catalog.pg_index n"
+				" WHERE pg_catalog.pg_index.indexrelid = %u"
+				"   AND n.indexrelid = 'reorg.index_%u'::regclass",
+				idx1, idx1);
+#endif
+			if (SPI_processed != 1)
+				elog(ERROR, "failed to update pg_index.indkey (%u rows updated)", SPI_processed);
+		}
 		CommandCounterIncrement();
 	}
 
 	/* swap names for toast tables and toast indexes */
-	if (reltoastrelid1 != InvalidOid)
+	if (reltoastrelid1 == InvalidOid)
+	{
+		if (reltoastidxid1 != InvalidOid ||
+			reltoastrelid2 != InvalidOid ||
+			reltoastidxid2 != InvalidOid)
+			elog(ERROR, "reorg_swap : unexpected toast relations (T1=%u, I1=%u, T2=%u, I2=%u",
+				reltoastrelid1, reltoastidxid1, reltoastrelid2, reltoastidxid2);
+		/* do nothing */
+	}
+	else if (reltoastrelid2 == InvalidOid)
+	{
+		char	name[NAMEDATALEN];
+
+		if (reltoastidxid1 == InvalidOid ||
+			reltoastidxid2 != InvalidOid)
+			elog(ERROR, "reorg_swap : unexpected toast relations (T1=%u, I1=%u, T2=%u, I2=%u",
+				reltoastrelid1, reltoastidxid1, reltoastrelid2, reltoastidxid2);
+
+		/* rename X to Y */
+		snprintf(name, NAMEDATALEN, "pg_toast_%u", oid2);
+		RenameRelationInternal(reltoastrelid1, name, PG_TOAST_NAMESPACE);
+		snprintf(name, NAMEDATALEN, "pg_toast_%u_index", oid2);
+		RenameRelationInternal(reltoastidxid1, name, PG_TOAST_NAMESPACE);
+		CommandCounterIncrement();
+	}
+	else if (reltoastrelid1 != InvalidOid)
 	{
 		char	name[NAMEDATALEN];
 		int		pid = getpid();
@@ -594,8 +876,12 @@ reorg_swap(PG_FUNCTION_ARGS)
 		CommandCounterIncrement();
 	}
 
+	/* adjust attribute numbers if the target table has dropped columns */
+	if (natts1 != natts2)
+		remove_dropped_columns_and_adjust_attnum(oid, natts1, natts2);
+
 	/* drop reorg trigger */
-	reorg_execf(
+	execute_with_format(
 		SPI_OK_UTILITY,
 		"DROP TRIGGER IF EXISTS z_reorg_trigger ON %s.%s CASCADE",
 		nspname, relname);
@@ -631,25 +917,38 @@ reorg_drop(PG_FUNCTION_ARGS)
 	 * drop reorg trigger: We have already dropped the trigger in normal
 	 * cases, but it can be left on error.
 	 */
-	reorg_execf(
+	execute_with_format(
 		SPI_OK_UTILITY,
 		"DROP TRIGGER IF EXISTS z_reorg_trigger ON %s.%s CASCADE",
 		nspname, relname);
 
+#if PG_VERSION_NUM < 80400
+	/* delete autovacuum settings */
+	execute_with_format(
+		SPI_OK_DELETE,
+		"DELETE FROM pg_catalog.pg_autovacuum v"
+		" USING pg_class c, pg_namespace n"
+		" WHERE relname IN ('log_%u', 'table_%u')"
+		"   AND n.nspname = 'reorg'"
+		"   AND c.relnamespace = n.oid"
+		"   AND v.vacrelid = c.oid",
+		oid, oid);
+#endif
+
 	/* drop log table */
-	reorg_execf(
+	execute_with_format(
 		SPI_OK_UTILITY,
 		"DROP TABLE IF EXISTS reorg.log_%u CASCADE",
 		oid);
 
 	/* drop temp table */
-	reorg_execf(
+	execute_with_format(
 		SPI_OK_UTILITY,
 		"DROP TABLE IF EXISTS reorg.table_%u CASCADE",
 		oid);
 
 	/* drop type for log table */
-	reorg_execf(
+	execute_with_format(
 		SPI_OK_UTILITY,
 		"DROP TYPE IF EXISTS reorg.pk_%u CASCADE",
 		oid);
@@ -659,6 +958,31 @@ reorg_drop(PG_FUNCTION_ARGS)
 	PG_RETURN_VOID();
 }
 
+Datum
+reorg_disable_autovacuum(PG_FUNCTION_ARGS)
+{
+	Oid			oid = PG_GETARG_OID(0);
+
+	/* connect to SPI manager */
+	reorg_init();
+
+#if PG_VERSION_NUM >= 80400
+	execute_with_format(
+		SPI_OK_UTILITY,
+		"ALTER TABLE %s SET (autovacuum_enabled = off)",
+		get_relation_name(oid));
+#else
+	execute_with_format(
+		SPI_OK_INSERT,
+		"INSERT INTO pg_catalog.pg_autovacuum VALUES (%u, false, -1, -1, -1, -1, -1, -1, -1, -1)",
+		oid);
+#endif
+
+	SPI_finish();
+
+	PG_RETURN_VOID();
+}
+
 /* init SPI */
 static void
 reorg_init(void)
@@ -678,41 +1002,6 @@ reorg_prepare(const char *src, int nargs, Oid *argtypes)
 	return plan;
 }
 
-/* execute prepared plan */
-static void
-reorg_execp(SPIPlanPtr plan, Datum *values, const char *nulls, int expected)
-{
-	int	ret = SPI_execute_plan(plan, values, nulls, false, 0);
-	if (ret != expected)
-		elog(ERROR, "pg_reorg: reorg_execp failed (code=%d, expected=%d)", ret, expected);
-}
-
-/* execute sql with format */
-static void
-reorg_execf(int expected, const char *format, ...)
-{
-	va_list			ap;
-	StringInfoData	sql;
-	int				ret;
-
-	initStringInfo(&sql);
-	va_start(ap, format);
-	appendStringInfoVA(&sql, format, ap);
-	va_end(ap);
-
-	if ((ret = SPI_exec(sql.data, 0)) != expected)
-		elog(ERROR, "pg_reorg: reorg_execf failed (sql=%s, code=%d, expected=%d)", sql.data, ret, expected);
-}
-
-/* execute a query */
-static void
-reorg_execd(const char *src, int nargs, Oid *argtypes, Datum *values, const char *nulls, int expected)
-{
-	int ret = SPI_execute_with_args(src, nargs, argtypes, values, nulls, expected == SPI_OK_SELECT, 0);
-	if (ret != expected)
-		elog(ERROR, "pg_reorg: reorg_execd failed (sql=%s, code=%d, expected=%d)", src, ret, expected);
-}
-
 static const char *
 get_quoted_relname(Oid oid)
 {
@@ -894,35 +1183,6 @@ swap_heap_or_index_files(Oid r1, Oid r2)
 
 extern PGDLLIMPORT bool allowSystemTableMods;
 
-static int
-SPI_execute_with_args(const char *src,
-					  int nargs, Oid *argtypes,
-					  Datum *values, const char *nulls,
-					  bool read_only, long tcount)
-{
-	SPIPlanPtr	plan;
-	int			ret;
-
-	plan = SPI_prepare(src, nargs, argtypes);
-	if (plan == NULL)
-		return SPI_result;
-	ret = SPI_execute_plan(plan, values, nulls, read_only, tcount);
-	SPI_freeplan(plan);
-	return ret;
-}
-
-static text *
-cstring_to_text(const char * s)
-{
-	int			len = strlen(s);
-	text	   *result = palloc(len + VARHDRSZ);
-
-	SET_VARSIZE(result, len + VARHDRSZ);
-	memcpy(VARDATA(result), s, len);
-
-	return result;
-}
-
 static void
 RenameRelationInternal(Oid myrelid, const char *newrelname, Oid namespaceId)
 {
